/***************************************************************************\
 * MMSOUND.c
 * (C) 94   Ralf Seidel
 *          Wlfrather Str. 45
 *          42105 Wuppertal
 *          email: seidel3@wrcs1.uni-wuppertal.de
 * and 
 * 			Nicole Greiber 
 * Set tabs to 3 to get a readable source.
\***************************************************************************/
#include <os2.h>
#define  INCL_OS2MM
#include <os2me.h>
#include <string.h>
#include <stdlib.h> 

#include "mmsound.h"

#ifdef DEBUG
/* Some dirty macros for debuging purposes. You may wonder about the usage
 * of fprintf. Redirect the output to a file and you will see the results. */
#include <stdio.h>
#ifdef __GCC__

/* This macro requires gcc */
#define PrintDebugInfo( format, arg... ) { \
   fprintf( stderr, "%3d: ", __LINE__ ); \
	fprintf( stderr, format, ## arg ); \
}

#else /* !__GCC__ */
#include <stdarg.h>

VOID PrintDebugInfo( PSZ format, ... )
{
	va_list arglist;

	va_start( arglist, format );
	vfprintf( stderr, format, arglist );
	va_end( arglist );
}

#endif /* __GCC__ */

#define PrintErrorInfo( rc ) { \
	CHAR errstr[512]; \
	mciGetErrorString( (rc), errstr, 512 ); \
	fprintf( stderr, "%3d: Error info: %s (0x%lX, %ld)\n", __LINE__, (errstr), (rc), (rc) ); \
}

#define DebugBeep( length ) DosBeep( 880, (length) );

#else /* !DEBUG */

#ifdef __GCC__
#define PrintDebugInfo( format, arg... ) ((void)0)
#else /* !__GCC__ */
VOID PrintDebugInfo( PSZ format, ... ) { }
#endif

#define PrintErrorInfo( rc ) ((void)0)
#define DebugBeep( length ) ((void)0)

#endif /* DEBUG */

static BOOL bDeviceOpen;

static USHORT usDeviceID;

static PLAYLIST aPlayList[2] = 
	{{ DATA_OPERATION, 0, 0, 0 },
	 { EXIT_OPERATION, 0, 0, 0 }};
	
/***********************************************************************\
 * FUNCTION AudioLoad
 * Opens a wav file, gets some information about the sampling
 * rate etc. and closes the file again.
 * The function returns MMIO_SUCCESS if the file was loaded correctly and
 * MMIO_ERROR if not.
\***********************************************************************/
ULONG AudioLoad( PSOUNDINFO pSndInfo )
{
	MMAUDIOHEADER mmAudioHeader;
	const WAVE_HEADER *pWaveHeader = &mmAudioHeader.mmXWAVHeader.WAVEHeader;
	LONG ulBytesRead = 0;
	HMMIO  hmmio;

	PrintDebugInfo( "AudioLoad: %s\n", pSndInfo->pszName );

	/* Initialize the header (set all entries to zero) */
	memset( &mmAudioHeader, 0, sizeof( mmAudioHeader ) );
	/* open the file pSndInfo->pszName */
	if ( ( hmmio = mmioOpen(pSndInfo->pszName, 0, MMIO_READ ) ) == NULLHANDLE ) {
		return MMIO_ERROR;
	}
	PrintDebugInfo( "mmioOpen Ok" );   

	/* get header */	 
	mmioGetHeader(hmmio, &mmAudioHeader, sizeof(mmAudioHeader), &ulBytesRead, 0, 0);
	/* get some infos about the PlayWavudio pSndInfo	*/
	pSndInfo->ulSamplesPerSec = pWaveHeader->ulSamplesPerSec;
	pSndInfo->usBitsPerSample = pWaveHeader->usBitsPerSample;
	pSndInfo->usChannels = pWaveHeader->usChannels;
	pSndInfo->ulSize = mmAudioHeader.mmXWAVHeader.XWAVHeaderInfo.ulAudioLengthInBytes;
	if ( (pSndInfo->pbData = malloc( pSndInfo->ulSize * sizeof( BYTE ))) == NULL ) {
		mmioClose(hmmio, 0);
		return MMIO_ERROR;
	}

	/* read sound data */
	ulBytesRead = 
	mmioRead( hmmio, pSndInfo->pbData, pSndInfo->ulSize );
	if ( ulBytesRead < 0 ) {
		free( pSndInfo->pbData );
		mmioClose(hmmio, 0);
		return MMIO_ERROR;
	}
	PrintDebugInfo( "mmioRead OK" ); 

	/* close pSndInfo file */
	mmioClose( hmmio, 0 );

	/* mark this sound as loaded */
	pSndInfo->bInMem = TRUE;

	return MMIO_SUCCESS;
}

/***********************************************************************\
 * FUNCTION AudioOpen
 * Opens the sound device and set the global variable usDeviceID
\***********************************************************************/
ULONG AudioOpen( const PSOUNDINFO pSndInfo )
{
	MCI_OPEN_PARMS mciOpenParms;
	MCI_WAVE_SET_PARMS mciWaveSetParms;
	ULONG rc;

	PrintDebugInfo( "AudioOpen\n" );

	if ( bDeviceOpen ) {
		PrintDebugInfo( "Device already opend\n" );
		return MCIERR_INVALID_MODE;
	}

	if ( !pSndInfo->bInMem ) {
		PrintDebugInfo( "Sound not loaded\n" );
		return MCIERR_OUT_OF_MEMORY;
	}

	/* open the device */
	mciOpenParms.hwndCallback   = NULLHANDLE;
	mciOpenParms.usDeviceID     = 
	mciOpenParms.usReserved0    = 0;
	mciOpenParms.pszDeviceType  = (PSZ)MCI_DEVTYPE_WAVEFORM_AUDIO_NAME;
	mciOpenParms.pszElementName = (PSZ)&aPlayList;
	rc =
	mciSendCommand( 0,
		MCI_OPEN,								/* open message */
		MCI_WAIT | MCI_OPEN_SHAREABLE |	/* message flags */
		MCI_OPEN_PLAYLIST,
		&mciOpenParms, 						/* parameters */
		0 );
	if ( (USHORT)rc != MCIERR_SUCCESS ) {
		PrintErrorInfo( rc );
		return rc;
	}
	PrintDebugInfo( "MCI_OPEN Ok\n" );
	usDeviceID = mciOpenParms.usDeviceID;

	/* set device parameters */
	mciWaveSetParms.hwndCallback	 = NULLHANDLE;
	mciWaveSetParms.ulSamplesPerSec = pSndInfo->ulSamplesPerSec;
	mciWaveSetParms.usBitsPerSample = pSndInfo->usBitsPerSample;
	mciWaveSetParms.usChannels     = pSndInfo->usChannels;
#if 0
	/* IBM writes that setting up the parameters has to be done step by step
	 * i.e. each parameter needs a single call. However it works if you
    * do all at once. */
	rc =
	mciSendCommand( usDeviceID, MCI_SET,
		MCI_WAVE_SET_SAMPLESPERSEC,
		&mciWaveSetParms, 0 );
	if ( (USHORT)rc != MCIERR_SUCCESS ) { 	
		PrintErrorInfo( rc );
		return rc;
	}
	rc =
	mciSendCommand( usDeviceID, MCI_SET,
		MCI_WAVE_SET_CHANNELS,
		&mciWaveSetParms, 0 );
	if ( (USHORT)rc != MCIERR_SUCCESS ) { 	
		PrintErrorInfo( rc );
		return rc;
	}
	rc =
	mciSendCommand( usDeviceID, MCI_SET,
		MCI_WAIT | MCI_WAVE_SET_BITSPERSAMPLE,
		&mciWaveSetParms, 0 );
	if ( (USHORT)rc != MCIERR_SUCCESS ) { 	
		PrintErrorInfo( rc );
		return rc;
	}
#else
	/* This method works on my computer without any problem. */
	rc =
	mciSendCommand( usDeviceID, MCI_SET,
		MCI_WAIT |
		MCI_WAVE_SET_BITSPERSAMPLE |
		MCI_WAVE_SET_CHANNELS |
		MCI_WAVE_SET_SAMPLESPERSEC,
		&mciWaveSetParms, 0 );
	if ( (USHORT)rc != MCIERR_SUCCESS ) { 		/* close the device */
		PrintErrorInfo( rc );
		return rc;
	}
#endif
	bDeviceOpen = TRUE;
	return MCIERR_SUCCESS;
}

/***********************************************************************\
 * FUNCTION AudioClose
 * Closes the sound device
\***********************************************************************/
ULONG AudioClose( VOID )
{
	MCI_GENERIC_PARMS mciGenericParms;

	bDeviceOpen = FALSE;
	PrintDebugInfo( "AudioClose\n" );
	mciGenericParms.hwndCallback = NULLHANDLE;
	return mciSendCommand( usDeviceID, MCI_CLOSE, MCI_WAIT, &mciGenericParms, 0 );
}


/***********************************************************************\
 * FUNCTION AudioPlay
 * Opens the sound device, plays the sound belonging to the pSndInfo
 * and closes the sound device.
\***********************************************************************/
ULONG AudioPlay( const PSOUNDINFO pSndInfo, const BOOL bCloseAfterwards )
{
	MCI_PLAY_PARMS mciPlayParms;
	ULONG rc;

	PrintDebugInfo( "AudioPlay: %s, %lX, %s, %lX, %hX, %hX\n" ,
		pSndInfo->pszName,
		pSndInfo->ulSize,
		pSndInfo->bInMem ? "TRUE" : "FALSE",
		pSndInfo->ulSamplesPerSec,
		pSndInfo->usBitsPerSample,
		pSndInfo->usChannels );

	if ( !bDeviceOpen &&
		  ( (USHORT)( rc = AudioOpen( pSndInfo ) ) != MCIERR_SUCCESS ) ) {
		return rc;
	}
	aPlayList[0].ulOp1 = (ULONG)pSndInfo->pbData;
	aPlayList[0].ulOp2 = pSndInfo->ulSize;
	aPlayList[0].ulOp3 = 0L;
	/* play the playlist
	 * Note: the `ulFrom = 0' parameter together with MCI_FROM option should be
	 * the default for the MCI_PLAY command. However I had some trouble if
	 * I omitted this parameter. */
	mciPlayParms.hwndCallback = NULLHANDLE;
	mciPlayParms.ulFrom = 0;
	rc = 
	mciSendCommand( usDeviceID, MCI_PLAY, MCI_WAIT | MCI_FROM, &mciPlayParms, 0);
	if ( (USHORT)rc != MCIERR_SUCCESS ) {
		PrintErrorInfo( rc );
	}
	if ( bCloseAfterwards ) {
		/* Close the sound device. */
		AudioClose();
	}
	return rc;
}

/***********************************************************************\
 * FUNCTION AudioPlayAsync
 * Like AudioPlay but doesn't initialize, doesn't wait for the sound
 * and doesn't close the sound device.
\***********************************************************************/
ULONG AudioPlayAsync(
	const HWND hwndCallback,
	const PSOUNDINFO pSndInfo,
	const BOOL bCloseAfterwards )
{
	MCI_PLAY_PARMS mciPlayParms;
	MCI_STATUS_PARMS mciStatusParms;
	USHORT usMsg;
	ULONG rc;

	PrintDebugInfo( "AudioPlayAsync: %s, %s\n" ,
		pSndInfo->pszName,
		pSndInfo->bInMem ? "TRUE" : "FALSE" );
	if ( !bDeviceOpen ) {
		rc =
		AudioOpen( pSndInfo );
		if ( rc != MCIERR_SUCCESS ) return rc;
	}
	/* Query the status of the audio device and stop playing if necessary */
	mciStatusParms.ulItem = MCI_STATUS_MODE;
	mciStatusParms.hwndCallback = NULLHANDLE;
	rc =
	mciSendCommand( usDeviceID, MCI_STATUS,
		MCI_WAIT | MCI_STATUS_ITEM, &mciStatusParms, 0 );
	if ( (USHORT) rc != MCIERR_SUCCESS ) {
		PrintErrorInfo( rc );
		return rc;
	}
	switch ( mciStatusParms.ulReturn ) {
		case MCI_MODE_PAUSE:
		case MCI_MODE_NOT_READY:
		case MCI_MODE_PLAY: {
			MCI_GENERIC_PARMS mciGenericParms = { NULLHANDLE };
			rc =
			mciSendCommand( usDeviceID, MCI_STOP, MCI_WAIT, &mciGenericParms, 0 );
			if ( (USHORT)rc != MCIERR_SUCCESS ) {
				PrintErrorInfo( rc );
				return rc;
			} else {
				PrintDebugInfo( "MCI_STOP Ok\n" );
			}
		}
		case MCI_MODE_STOP:
			break;
		default:
			PrintDebugInfo( "Found status mode: %ld\n", mciStatusParms.ulReturn  );
			return MCIERR_INVALID_MODE;
	} /* end switch ulReturn */

	/* Set play list parameters to the new sound data. */
	aPlayList[0].ulOp1 = (ULONG)pSndInfo->pbData;
	aPlayList[0].ulOp2 = pSndInfo->ulSize;
	aPlayList[0].ulOp3 = 0;

	/* play the playlist */
	if ( bCloseAfterwards ) {
		usMsg = MCI_NOTIFY | MCI_FROM;
		mciPlayParms.hwndCallback = hwndCallback;
	} else {
		usMsg = MCI_FROM;
		mciPlayParms.hwndCallback = NULLHANDLE;
	}
	mciPlayParms.ulFrom = 0;
	rc = 
	mciSendCommand( usDeviceID, MCI_PLAY, usMsg, &mciPlayParms, 0 );
	if ( (USHORT)rc != MCIERR_SUCCESS ) {
		PrintErrorInfo( rc );
	} else {
		PrintDebugInfo( "MCI_PLAY Ok\n" );
	}
	return rc;
}

/***********************************************************************\
 * FUNCTION AudioFree
 * Frees the memory that has been allocated for the sound.
\***********************************************************************/
BOOL AudioFree( PSOUNDINFO pSndInfo )
{
	PrintDebugInfo( "AudioFree: %s, %lX, %s, %lX, %hX, %hX\n" ,
		pSndInfo->pszName,
		pSndInfo->ulSize,
		pSndInfo->bInMem ? "TRUE" : "FALSE",
		pSndInfo->ulSamplesPerSec,
		pSndInfo->usBitsPerSample,
		pSndInfo->usChannels );
	if ( !pSndInfo ) return FALSE;
	pSndInfo->bInMem = FALSE;
	free( pSndInfo->pbData );
	return TRUE;
}

/***********************************************************************\
 * FUNCTION AudioPlayFile
 * Plays a *.wav or *.voc file.
\***********************************************************************/
BOOL AudioPlayFile( const PSZ pszFileName )
{
	SOUNDINFO SndInfo;
	ULONG rc;

	SndInfo.pszName = pszFileName;
	if ( AudioLoad( &SndInfo ) != MMIO_SUCCESS ) return FALSE;
	rc = AudioPlay( &SndInfo, TRUE );
	AudioFree( &SndInfo );
	return ( (USHORT)rc == MCIERR_SUCCESS );
}


/***********************************************************************\
 * FUNCTION AudioWndProc
 * Checks if msg is a MM_MNCINOTIFY message. If true and if usMessage
 * (= SHORT2FROMMP( mp2 ) ) is equal MCI_PLAY it closes the sound device
 * usDeviceID ( == SHORT1FROMMP( mp2 )). In this case TRUE is returned to
 * indicate that the message has been handled. FALSE is returned if
 * this function can't handle the message.
\***********************************************************************/
BOOL AudioWndProc(
	const HWND hwnd,
	const ULONG msg,
	const MPARAM mp1,
	const MPARAM mp2 )
{
	switch ( msg ) {
		case MM_MCINOTIFY: {
			const USHORT usMessage = SHORT2FROMMP( mp2 );
			if ( usMessage == MCI_PLAY ) {
				AudioClose();
				return TRUE;
			}
			break;
		} /* end case MM_MCINOTIFY */
	} /* end switch msg */
	return FALSE;
}

