/*************************************************************************\
 * shootem.c
 * (C) 94   Ralf Seidel
 *          Wlfrather Str. 45
 *          42105 Wuppertal
 *          email: seidel3@wrcs1.uni-wuppertal.de
 *
 * Set tabs to 3 to get a readable source.
\*************************************************************************/
/* include files, macros and definitions */
#define INCL_DEV
#define INCL_DOSPROCESS
#define INCL_DOSMEMMGR
#define INCL_WINMENUS
#define INCL_WINSYS
#define INCL_WINHELP
#define INCL_WININPUT
#define INCL_WINFRAMEMGR
#define INCL_WINWINDOWMGR
#define INCL_WINMESSAGEMGR
#define INCL_WINDIALOGS
#define INCL_WINSHELLDATA
#define INCL_WINTIMER
#define INCL_WINRECTANGLES
#define INCL_WINPOINTERS
#define INCL_GPIPRIMITIVES
#define INCL_GPIBITMAPS
#include <os2.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <getopt.h>
#include <time.h>

#include "loadbmp.h"
#include "errmsgs.h"
#include "mmsound.h"
#include "resdefs.h"

/* type definition and global variables used inside this module. */

typedef struct {
	HBITMAP Mask;
	HBITMAP Image;
} TRANSPARENT_BITMAP, *PTRANSPARENT_BITMAP;


typedef struct {
	INT dummy;
} PROFILEDATA, *PPROFILEDATA;

static const struct {
	PSZ pszSettings;
	PSZ pszWindowSize;
} PrfKeyNames = { "Settings", "Size" };

typedef struct {
	PSZ pszProgName;		/* Programname (*.exe) */
	PSZ pszProgPath;		/* path of the executable (including drive) */
	PSZ pszAppTitle;		/* Name of this application */
	PSZ pszHelpName;		/* Name of the related help file */
	PSZ pszIniName;		/* Name of the ini file */
	HAB hab;					/* Anchor block of the main thread */
	HWND hwndMain;			/* window handle of the main window */
	HWND hwndHelp;			/* window of the help facility */
	/* Variable settings */
	UINT fSaveSettings : 1;
	UINT fRestart: 1;
	UINT fPreload: 1;
	UINT fSound: 1;
	ULONG ulMinSleep;
	ULONG ulMaxSleep;
	ULONG ulFrameRate;
} PROGINFO, *PPROGINFO;

static PROGINFO ProgInfo;

typedef struct {
	ULONG ulStructSize;
	HWND hwndCaller;
	PSOUNDINFO apSndInfo[1];
} SNDLOAD_THREADDATA, *PSNDLOAD_THREADDATA;

/* A user defined window message */
#define MYM_ENDSNDLOAD (WM_USER)
#define MYM_STARTSLEEPTIMER (WM_USER + 1)


#define PULL_BMPS			6
#define HIT_BMPS			5
#define PULL_PICS			(PULL_BMPS)
#define HIT_PICS			(HIT_BMPS)

#define MIN_FRAMERATE	100

#define LOAD_BMPS_FROM_FILE 0
#define LOAD_SNDS_FROM_FILE 0

#if LOAD_BMPS_FROM_FILE

BOOL LoadTransparentBitmap(
	const HPS hpsMem, const PSZ pszFileName, PTRANSPARENT_BITMAP ptb );

static const PSZ apszPullFiles[PULL_BMPS] =
	{ "pull0.bmp", "pull1.bmp", "pull3.bmp", "pull4.bmp", "pull5.bmp", "pull6.bmp" };

static const PSZ apszHitFiles[HIT_BMPS] =
	{ "hit0.bmp", "hit1.bmp", "hit2.bmp", "hit3.bmp", "hit4.bmp" };

#define LoadPullImage( hpsMem, i, ptbmp ) \
	LoadTransparentBitmap( (hpsMem), apszPullFiles[i], (ptbmp) )
#define LoadHitImage( hpsMem, i, ptbmp ) \
	LoadTransparentBitmap( (hpsMem), apszHitFiles[i], (ptbmp) )

#else

BOOL LoadTransparentBitmap(
	const HPS hpsMem, const ULONG ulBmpId, PTRANSPARENT_BITMAP ptb );

static const ULONG alPullIds[PULL_BMPS] = {
	IDBMP_PULL0, IDBMP_PULL1, IDBMP_PULL2,
	IDBMP_PULL3, IDBMP_PULL4, IDBMP_PULL5 };

static const ULONG alHitIds[HIT_BMPS] = {
	IDBMP_HIT0, IDBMP_HIT1, IDBMP_HIT2,
	IDBMP_HIT3, IDBMP_HIT4 };

#define LoadPullImage( hpsMem, i, ptbmp ) \
	LoadTransparentBitmap( (hpsMem), alPullIds[i], (ptbmp) )
#define LoadHitImage( hpsMem, i, ptbmp ) \
	LoadTransparentBitmap( (hpsMem), alHitIds[i], (ptbmp) )

#endif

/* prototypen of the functions defined by the program */

MRESULT EXPENTRY MainClientProc(	const HWND hwnd, const ULONG msg,
	const MPARAM mp1, const MPARAM mp2 );

/* Some dirty macros: (debugging aids)
 * Not all are used here. (I just coppied them from an other application.) */

#define ErrMsgRunLib( msg ) _ErrMsgRunLib( (msg), __FILE__, __LINE__ )
#define ErrMsgPM( hab, msg ) _ErrMsgPM( (hab), (msg), __FILE__, __LINE__ );	
#define ErrMsgKernel( errcode, msg ) _ErrMsgKernel( (errcode), (msg), __FILE__, __LINE__ );

#define Fatal( msg ) \
{ \
	CHAR fullmsg[512]; sprintf( fullmsg, "%s\nFile: %s Line: %d", msg, __FILE__, __LINE__ ); \
	WinMessageBox( HWND_DESKTOP, NULLHANDLE, fullmsg, "Error", 0, MB_ERROR | MB_OK ); \
	WinSendMsg( ProgInfo.hwndMain, WM_CLOSE, (MPARAM)0, (MPARAM)0 ); \
}	
#define KernelFatal( errcode, msg ) \
{ \
	ErrMsgKernel( (errcode), (msg) );	\
	WinSendMsg( ProgInfo.hwndMain, WM_CLOSE, (MPARAM)0, (MPARAM)0 ); \
}
#define PMFatal( hab, msg ) \
{ \
	ErrMsgPM( hab, (msg) );	\
	WinSendMsg( ProgInfo.hwndMain, WM_CLOSE, (MPARAM)0, (MPARAM)0 ); \
}
#define RunLibFatal( msg ) \
{ \
	ErrMsgRunLib( (msg) );	\
	WinSendMsg( ProgInfo.hwndMain, WM_CLOSE, (MPARAM)0, (MPARAM)0 ); \
}

#define Xmalloc( ptr, size, action_if_failed ) \
	if ( !((ptr) = (PVOID)malloc( (size_t)(size) ) ) ) { \
		ErrMsgRunLib( "malloc" ); \
		WinSendMsg( ProgInfo.hwndMain, WM_CLOSE, (MPARAM)0, (MPARAM)0 ); \
		action_if_failed; \
	}
#define Xcalloc( ptr, size, action_if_failed ) \
	if ( !((ptr) = (PVOID)calloc( (size_t)(size), 1 ) ) ) { \
		ErrMsgRunLib( "calloc" ); \
		WinSendMsg( ProgInfo.hwndMain, WM_CLOSE, (MPARAM)0, (MPARAM)0 ); \
		action_if_failed; \
	}
#define Xrealloc( ptr, size, action_if_failed ) \
	if ( !((ptr) = (PVOID)realloc( (PVOID)p, (size_t)(size) ) ) ) { \
		ErrMsgRunLib( "realloc" ); \
		WinSendMsg( ProgInfo.hwndMain, WM_CLOSE, (MPARAM)0, (MPARAM)0 ); \
		action_if_failed; \
	}
#define Xalloca( ptr, size, action_if_failed ) \
	if ( !((ptr) = (PVOID)alloca( size ) ) ) { \
		ErrMsgRunLib( "alloca" ); \
		WinSendMsg( ProgInfo.hwndMain, WM_CLOSE, (MPARAM)0, (MPARAM)0 ); \
		action_if_failed; \
	} 

/*************************************************************************\
 * PROCEDURE SetProgInfoDefaults
 * Initializes the variable fields of the ProgInfo structure if no valid
 * data is saved in the profile.
\*************************************************************************/
VOID SetProgInfoDefaults( VOID )
{
	ProgInfo.fSaveSettings = 1;
	ProgInfo.fRestart = 0;
	ProgInfo.fPreload = 0;
	ProgInfo.fSound = 1;
	ProgInfo.ulMinSleep = 10000;
	ProgInfo.ulMaxSleep = 0;
	ProgInfo.ulFrameRate = 2 * MIN_FRAMERATE;
	return;
}
	
/*************************************************************************\
 * FUNCTION StoreSettings
 * Saves current settings of ProgInfo.
\*************************************************************************/
VOID StoreSettings( VOID )
{
#if 0
	/* I don't use ini file settings. This function is supplied for
	 * possible future enhancements only */
	HINI hini;
	PROFILEDATA PrfData;
	PSZ pszIniName;
	LONG lNumWrite;
	
	Xalloca( pszIniName, strlen( ProgInfo.pszIniName ) + 1, return ) ;
	strcpy( pszIniName, ProgInfo.pszIniName );
	if ( !(hini = PrfOpenProfile( ProgInfo.hab, pszIniName )) ) 
		return;

	lNumWrite = sizeof( PROFILEDATA );
	PrfWriteProfileData(
		hini,	ProgInfo.pszAppTitle,
		PrfKeyNames.pszSettings, &PrfData, lNumWrite );
	PrfCloseProfile( hini );
	return;
#endif
}

/*************************************************************************\
 * FUNCTION RestoreSettings
 * Initialize the variable fields of the ProgInfo structure.
 * Returns TRUE if profiledata was successful retrieved and FALSE
 * otherwise.
\*************************************************************************/
BOOL RestoreSettings( PPROFILEDATA pPrfData )
{
#if 1
	/* I don't use ini file settings */
	return FALSE;
#else
	BOOL bResult = TRUE;
	HINI hini;
	PSZ pszIniName;
	LONG lNumRead;

	Xalloca( pszIniName, strlen( ProgInfo.pszIniName ) + 1, return FALSE );
	strcpy( pszIniName, ProgInfo.pszIniName );
	if ( !(hini = PrfOpenProfile( ProgInfo.hab, pszIniName )) ) 
		return FALSE;
	lNumRead = sizeof( PROFILEDATA );

	bResult =
	PrfQueryProfileData(
		hini,	ProgInfo.pszAppTitle,
		PrfKeyNames.pszSettings, pPrfData, &lNumRead )
	&& (lNumRead == sizeof( PROFILEDATA ));
		
	PrfCloseProfile( hini );
	return bResult;
#endif
}

/*************************************************************************\
 * PROCEDURE ShowHelp
\*************************************************************************/
VOID ShowHelp( VOID )
{
	CHAR msg[256];
	CHAR title[128];
	sprintf( msg,
		"-q : quiet\n"
		"-p : Preload bitmaps\n"
		"-r : Respawn\n"
		"-m n : Minmal sleep time\n" 
		"-M n : Maximal sleep time\n"
		"-s n : Opponent speed (0 - 9)" );
	sprintf( title,
		"Usage: %s -p -r -m n -M n -s n", ProgInfo.pszProgName );
	WinMessageBox( HWND_DESKTOP, NULLHANDLE,
		msg, title, 0,	MB_INFORMATION | MB_OK | MB_MOVEABLE );
	return;
}


/*************************************************************************\
 * PROCEDURE InitProgInfo
 * Initialize the fields of the ProgInfo structure.
\*************************************************************************/
static BOOL InitProgInfo( INT argc, CHAR *argv[], PSZ pszBase )
{
	CHAR drive[_MAX_DRIVE];
	CHAR dir[_MAX_DIR];
	CHAR Name[_MAX_FNAME];
	CHAR ext[_MAX_EXT];
	PROFILEDATA PrfData;
	INT iBaseLen1 = strlen( pszBase ) + 1;
	INT c;
	BOOL bOptErr;

	Xmalloc( ProgInfo.pszAppTitle, iBaseLen1, return TRUE );
	strcpy( ProgInfo.pszAppTitle, pszBase );
	
	_splitpath( argv[0], drive, dir, Name, ext );
	Xmalloc( ProgInfo.pszProgPath, strlen( drive ) + strlen( dir ) + 1, return TRUE );
	strcpy( ProgInfo.pszProgPath, drive );
	strcat( ProgInfo.pszProgPath, dir );
	Xmalloc( ProgInfo.pszProgName, strlen( Name ) + strlen( ext ) + 1, return TRUE );
	strcpy( ProgInfo.pszProgName, Name );
	strcat( ProgInfo.pszProgName, ext );
	strcpy( ext, ".hlp" );
	Xmalloc( ProgInfo.pszHelpName, iBaseLen1 + strlen( ext ) , return TRUE );
	strcpy( ProgInfo.pszHelpName, pszBase );
	strcat( ProgInfo.pszHelpName, ext );
	strcpy( ext, ".ini" );
	Xmalloc( ProgInfo.pszIniName, iBaseLen1 + strlen( ext ) , return TRUE );
	strcpy( ProgInfo.pszIniName, pszBase );
	strcat( ProgInfo.pszIniName, ext );

	if ( RestoreSettings( &PrfData ) ) {
		ProgInfo.fSaveSettings = 0;
	} else {
		SetProgInfoDefaults();
	}

	/* Parse command line options */
	bOptErr = FALSE;
	opterr = 0;
#ifdef __EMX__
	optswchar = "-/";	/* Allow either case of switch characters */
#endif
	while ( !bOptErr && ( c = getopt( argc, argv, "m:M:pqrs:" ) ) != EOF ) {
		switch ( c ) {
			case 'm':
				ProgInfo.ulMinSleep = 1000 * strtoul( optarg, NULL, 0 );
				break;
			case 'M':
				ProgInfo.ulMaxSleep = 1000 * strtoul( optarg, NULL, 0 );
				break;
			case 'q':
				ProgInfo.fSound = 0;
				break;
			case 'p':
				ProgInfo.fPreload = 1;
				break;
			case 'r':
				ProgInfo.fRestart = 1;
				break;
			case 's':
				/* Only examine the first character of the command argument */
				if ( !optarg || !isdigit( optarg[0] ) ) {
					bOptErr = TRUE;
				} else {
					ProgInfo.ulFrameRate = MIN_FRAMERATE + MIN_FRAMERATE * (optarg[0] - '0') / 5;
				}
				break;
			default:
				bOptErr = TRUE;
		} /* end switch c */
	} /* end while getopt */
	if ( !ProgInfo.ulMaxSleep ) {
		/* maximal sleep period wasn't set */
		ProgInfo.ulMaxSleep = ProgInfo.ulMinSleep + 1;
	} else if ( ProgInfo.ulMaxSleep <= ProgInfo.ulMinSleep ) {
		bOptErr = TRUE;
	}
	/********************************************************************\
	 * Remark: the timer values are limited because of an OS/2 bug in
	 * WinStartTimer. This functions doesn't accept any timeout values
	 * greater than 0xFFFF.
	\********************************************************************/
	if ( bOptErr ) {
		ShowHelp();
		return FALSE;
	}
	return TRUE;
}

#if 0	
/*************************************************************************\
 * FUNCTION CreateBitmapMask
 * This function creates a mask for a bitmap.
\*************************************************************************/
HBITMAP CreateBitmapMask( HAB hab, HBITMAP hbmpImage )
{
	HPS hpsMem;
	HBITMAP hbmpMask;
	BITMAPINFOHEADER2 bmihImage;
	BITMAPINFOHEADER2 bmihMask;
	ULONG cbBuffer;
	ULONG cbBitmapInfo;
	PBYTE pbBuffer;
	PBITMAPINFO2 pbmi;
	USHORT usBits;
	LONG lResult;
	INT i;

	/* Create a presentations space associated with a memory device */
	if ( !(hpsMem = CreateMemPS( hab )) ) return NULLHANDLE;

	/* Query bitmap information of the image bitmap */
	bmihImage.cbFix = sizeof( BITMAPINFOHEADER2 );
	if ( !GpiQueryBitmapInfoHeader( hbmpImage, &bmihImage ) ) return NULLHANDLE;
	usBits = bmihImage.cBitCount * bmihImage.cPlanes;
	
	/* Get some memory for the GpiQueryBitmapBits return values */
	cbBuffer = ((( bmihImage.cBitCount * bmihImage.cx ) + 31 ) / 32 )
		* 4 * bmihImage.cy * bmihImage.cPlanes;
	if ( usBits != 24 ) {
		cbBitmapInfo = sizeof( BITMAPINFO2 ) + (sizeof( RGB ) * ( 1 << usBits ));
	} else {
		cbBitmapInfo = sizeof( BITMAPINFO2 );
	}
   /* Doesn't work... Why ?
	Xalloca( pbBuffer, cbBuffer, return (NULLHANDLE) );
	Xalloca( pbmi, cbBitmapInfo, return (NULLHANDLE) );
	*/

	Xmalloc( pbBuffer, cbBuffer, return (NULLHANDLE) );
	Xmalloc( pbmi, cbBitmapInfo, return (NULLHANDLE) );

	/* Copy the information of the image bitmap */
	memcpy( pbmi, &bmihImage, sizeof( BITMAPINFOHEADER2 ) );

	/* Select the image bitmap to query its bits */
	GpiSetBitmap( hpsMem, hbmpImage );
	lResult = 
	GpiQueryBitmapBits( hpsMem, 0L, bmihImage.cy, pbBuffer, pbmi );
	if ( lResult == GPI_ALTERROR ) {
		GpiSetBitmap( hpsMem, NULLHANDLE );
		DestroyMemPS( hpsMem );
		free( pBuffer );
		free( pbmi );
		return NULLHANDLE;
	}

	memcpy( &bmihMask, &bmihImage, sizeof( BITMAPINFOHEADER2 ) );
	bmihMask.cPlanes = 1;
	bmihMask.cBitCount = 1;

	if ( usBits != 24 ) {
		const RGB2 rgb2White = { 255, 255, 255, 0 };
		const PBITMAPINFO2 pbmpi2 = (PBITMAPINFO2)pbmpih2;
		pbmpi2->argbColor[0] = rgb2White;
		memset( &pbmpi2->argbColor[1], 0, ((1<<usBits ) - 1) * sizeof( RGB2 ) );
	}

	GpiSetColor( hpsMem, CLR_WHITE );
	GpiSetBackColor( hpsMem, CLR_BLACK );
	hbmpMask =
	GpiCreateBitmap( hpsMem, &bmihMask, CBM_INIT, pbBuffer, pbmi );

	free( pBuffer );
	free( pbmi );
	GpiSetBitmap( hpsMem, NULLHANDLE );
	DestroyMemPS( hpsMem );
	return hbmpMask;
}
#endif

/*************************************************************************\
 * FUNCTION GetWindowBitmap
 * This function creates a bitmap of the current window contents.
\*************************************************************************/
HBITMAP GetWindowBitmap( HWND hwnd, PRECTL prect )
{
	HPS hpsWnd = NULLHANDLE;
	HDC hdcWnd = NULLHANDLE;
	HPS hpsMem = NULLHANDLE;
	HBITMAP hbmpReturn = NULLHANDLE;
	BITMAPINFOHEADER2 bmpih2;
	LONG alCaps[2];
	POINTL aptl[4];
	RECTL rect;

	hpsWnd = WinGetPS( hwnd );
	if ( !hpsWnd ) return NULLHANDLE;

	/* Query the number of planes and pixel of the video device */
	hdcWnd = GpiQueryDevice( hpsWnd );
	DevQueryCaps( hdcWnd, CAPS_COLOR_PLANES, 2, alCaps );

	/* Determine the size of the window if no rectangle is submitted. */
	if ( !prect ) {
		WinQueryWindowRect( hwnd, &rect );
		prect = &rect;
	}

	/* Initialies the BITMAPINFOHEADER2 structure for GpiCreateBitmap */
#if 1
	/* PM doesn't need the full bitmap information */
	bmpih2.cbFix = 16L;
 	bmpih2.cx = prect->xRight - prect->xLeft;
	bmpih2.cy = prect->yTop - prect->yBottom;
	bmpih2.cPlanes = alCaps[0]; 
	bmpih2.cBitCount = alCaps[1];
#else
	bmpih2.cbFix = sizeof( BITMAPINFOHEADER2 );
 	bmpih2.cx = prect->xRight - prect->xLeft;
	bmpih2.cy = prect->yTop - prect->yBottom;
	bmpih2.cPlanes = alCaps[0]; 
	bmpih2.cBitCount = alCaps[1];
	bmpih2.ulCompression = BCA_UNCOMP;
	bmpih2.cbImage = 0;				/* Bitmap is uncompress -> this field
											 * can be zero */
	bmpih2.cxResolution = 640; 	/* set to standard VGA Resolution */ 
	bmpih2.cyResolution = 480;
	bmpih2.cclrUsed = 0; 			/* Assume all colors are used */
	bmpih2.cclrImportant = 0;
	bmpih2.usUnits = BRU_METRIC;
	bmpih2.usReserved = 0;
	bmpih2.usRecording = BRA_BOTTOMUP;
	bmpih2.usRendering = BRH_NOTHALFTONED;
	bmpih2.cSize1 = 0;
	bmpih2.cSize2 = 0;
	bmpih2.ulColorEncoding = BCE_RGB;
	bmpih2.ulIdentifier = 0;
#endif

	/* Create a memory PS to hold the bitmap data */
	hpsMem = CreateMemPS( WinQueryAnchorBlock( hwnd ) );
	if ( !hpsMem ) goto clean;
	hbmpReturn = GpiCreateBitmap( hpsMem, &bmpih2, 0, NULL, NULL );
	if ( !hbmpReturn ) goto clean;
	/* Target rectangle */
	aptl[0].x = 0;
	aptl[0].y = 0;
	aptl[1].x =	prect->xRight - prect->xLeft + 1;
	aptl[1].y = prect->yTop - prect->yBottom + 1;
	/* Source rectangle */
	aptl[2].x =	prect->xLeft;
	aptl[2].y = prect->yBottom;
	aptl[3].x =	prect->xRight + 1;
	aptl[3].y = prect->yTop + 1;
	GpiSetBitmap( hpsMem, hbmpReturn );
	GpiBitBlt( hpsMem, hpsWnd, 4, aptl, ROP_SRCCOPY, BBO_IGNORE );
	GpiSetBitmap( hpsMem, NULLHANDLE );

clean:
	if ( hpsWnd ) WinReleasePS( hpsWnd );
	if ( hpsMem ) DestroyMemPS( hpsMem );
	return hbmpReturn;
}
	
/*************************************************************************\
 * FUNCTION LoadTransparentBitmap
 * Load bmp file and create the image and mask bitmaps.
\*************************************************************************/
#if LOAD_BMPS_FROM_FILE
BOOL LoadTransparentBitmap(
	const HPS hpsMem, const PSZ pszFileName, PTRANSPARENT_BITMAP ptb )
#else
BOOL LoadTransparentBitmap(
	const HPS hpsMem, const ULONG ulBmpId, PTRANSPARENT_BITMAP ptb )
#endif
{
   PBITMAPINFOHEADER2 pbmpih2; 	/* Address any info headers        */
   BITMAPINFOHEADER2 bmpih2Mask;
   PBYTE pBuffer;						/* Holds the bitmap data (pixels)  */
   USHORT usBits;
	BOOL bResult = FALSE;

#if LOAD_BMPS_FROM_FILE
	if ( !_LoadBitmap( pszFileName, NULLHANDLE, 0L, &pbmpih2, &pBuffer ) )
		return FALSE;
#else	
	if ( !_LoadBitmap( NULL, NULLHANDLE, ulBmpId, &pbmpih2, &pBuffer ) )
		return FALSE;
#endif

   ptb->Image =
   GpiCreateBitmap(
      hpsMem,							/* Presentation-space handle */
      pbmpih2,             		/* Address of structure for format data */
      CBM_INIT,              		/* Options: create bitmap bits now */
      pBuffer,                	/* Address of buffer of image data */
      (PBITMAPINFO2)pbmpih2);    /* Address of structure for color and format */
   if ( !ptb->Image ) goto clean;

	memcpy( &bmpih2Mask, pbmpih2, pbmpih2->cbFix );
  	usBits = bmpih2Mask.cBitCount * bmpih2Mask.cPlanes;
	if ( usBits != 24 ) {
		/* Set first palette entry to white (background) and all others to
		 * black.
		 * Note: At the first try I accessed memory at location pbmpi2->argbColor
		 * which is a fault because the palette information starts immediatly
		 * at (PBYTE)pbmpi2 + pbmpi2->cbFix */
#if 1
		const RGB2 rgb2White = { 255, 255, 255, 0 };
		const PBITMAPINFO2 pbmpi2 = (PBITMAPINFO2)pbmpih2;
		PRGB2 pPal = (PRGB2)( (PBYTE)pbmpi2 + pbmpi2->cbFix );

		*pPal = rgb2White;
		pPal++;
		memset( pPal, 0, ((1<<usBits ) - 1) * sizeof( RGB2 ) );
		/* The last lines should be equal to: */
#else	
		const RGB2 rgb2White = { 255, 255, 255, 0 };
		const RGB2 rgb2Black = { 0, 0, 0, 0 };
		const PBITMAPINFO2 pbmpi2 = (PBITMAPINFO2)pbmpih2;
		PRGB2 pPal = (PRGB2)( (PBYTE)pbmpi2 + pbmpi2->cbFix );
		INT i;

		*pPal = rgb2White;
		for ( i = 1; i < ( 1 << usBits ); i++ ) {
			++pPal;
			*pPal = rgb2Black;
		}
#endif
	} else {
		/* TODO */
		/* Don't know what happens if we handle true color bitmaps */
	}
	bmpih2Mask.cBitCount = 1;
	bmpih2Mask.cPlanes = 1;
	/* Just to be sure */
	GpiSetColor( hpsMem, CLR_WHITE );
	GpiSetBackColor( hpsMem, CLR_BLACK );
	ptb->Mask =
	GpiCreateBitmap(
		hpsMem,
		&bmpih2Mask,
		CBM_INIT,
		pBuffer,
		(PBITMAPINFO2)pbmpih2 );

	bResult = (ptb->Mask != NULLHANDLE);
clean:
   /* release memory allocated by _LoadBitmap */
	DosFreeMem( pBuffer );
	DosFreeMem( pbmpih2 );

   return bResult;  
}

/*************************************************************************\
 * PROCEDURE LoadSounds
 * This procedure is called as a thread to load all sound data.
\*************************************************************************/
VOID LoadSounds( PVOID pvParm )
{
	const PSNDLOAD_THREADDATA pSndData = (PSNDLOAD_THREADDATA)pvParm;
	const INT n =
		( pSndData->ulStructSize - (INT)( (ULONG)&pSndData->apSndInfo[0] - (ULONG)pSndData ) )
		/ sizeof( PSOUNDINFO );
	const HAB habThread = WinInitialize( 0 );
	const HMQ hmqThread = WinCreateMsgQueue( habThread, 0 );
	INT i;
#ifdef DEBUG
	DosEnterCritSec();
	fprintf( stderr, "LoadSounds call to load %d files: ", n);
	for ( i = 0; i < n; i++ ) {
		fprintf( stderr, "%s ", pSndData->apSndInfo[i]->pszName );
	}
	fputc( '\n', stderr );
	DosExitCritSec();
#endif

	if ( (habThread == NULLHANDLE) || (hmqThread == NULLHANDLE) ) {
		if ( hmqThread ) WinDestroyMsgQueue( hmqThread );
		if ( habThread ) WinTerminate( habThread );
		return;
	}
	for ( i = 0; i < n; i++ ) {
		/* Ignore Errors */
		if ( AudioLoad( pSndData->apSndInfo[i] ) /* != MMIO_SUCCESS */ ) {
			pSndData->apSndInfo[i]->bInMem = FALSE;
		}
	}
	WinSendMsg( pSndData->hwndCaller, MYM_ENDSNDLOAD, (MPARAM)pSndData, (MPARAM)0 );
	WinDestroyMsgQueue( hmqThread );
	WinTerminate( habThread );
	/* Someone wrote that the following statement is a good programming style.
	 * I never understood why... */
	DosEnterCritSec();
}


/*************************************************************************\
 * Client-Window-procedure of the main window
 * Here all the message handling takes place.
\*************************************************************************/
MRESULT EXPENTRY MainClientProc(
	const HWND hwnd,
  	const ULONG msg,
	const MPARAM mp1,
	const MPARAM mp2 )
{
	/* Some sounds */
	static SOUNDINFO sndHaltStop = { "haltstop.wav", FALSE };
	static SOUNDINFO sndGudenDag = { "gudendag.wav", FALSE };
	static SOUNDINFO sndBang = { "bang1.wav", FALSE };
	static SOUNDINFO sndArrgh = { "arrgh.wav", FALSE };
	static HPS hpsBuffer = NULLHANDLE;
	/* The current active bitmap */
  	static PTRANSPARENT_BITMAP ptbmpCurrent = NULL;
  	/* A bitmap to save the background of our window */
	static HBITMAP hbmpBkg = NULLHANDLE;
	/* We need this presentation space to avoid flicker during the painting
	 * of bitmaps */
	static HBITMAP hbmpBuffer = NULLHANDLE;

	static TRANSPARENT_BITMAP hbmpPull[PULL_PICS];
	static TRANSPARENT_BITMAP hbmpHit[HIT_PICS];
	static BOOL bInPullPhase = FALSE;
	static BOOL bRestart;
	static HPOINTER hptrXhair;	/* Handle for a pointer used as a crosshair */
	static SHORT sSleep2Ct;
	

	/* some timer identifiers */
	enum {
		tidSleep = 1,
		tidMinutes, /* This timer doesn't really count minutes but it counts
						 * 0xFFFF = 65535 milli seconds. */
		tidPull,
		tidHit,
		tidFired,
		tidWait };
	static INT iBmpIdxPull = 0;
	static INT iBmpIdxHit = 0;

#ifdef DEBUG
	fprintf( stderr, "msg: %8lX, %8lX, %8lX\n",
		(ULONG)msg, (ULONG)mp1, (ULONG)mp2 );
#endif
	switch ( msg ) {
		case WM_PAINT: {
			if ( ptbmpCurrent != NULL ) {
				RECTL rect;
				HPS hps = WinBeginPaint( hwnd, 0, &rect );
				/* We need this variable to determine the bitmap size */
				BITMAPINFOHEADER bmpInfo;
				/* Select bitmap area which has to be drawn (updated): */
				POINTL aptUpdate[4] = {
					rect.xLeft, 		rect.yBottom,
					rect.xRight, 		rect.yTop, 
					rect.xLeft,			rect.yBottom,
					rect.xRight + 1,	rect.yTop + 1 };
				POINTL aptBmp[4];

				/**************************************************************\
				 * Note: Following information is taken from the OS/2 toolkit
				 * reference for GpiWCBitBlt
				 * ...
				 * A rectangle is specified in device coordinates for the
				 * source bit map, and one in world coordinates for the
				 * target presentation space. The source rectangle is
				 * noninclusive; the left and lower boundaries in device
				 * space are included, but not the right and upper
				 * boundaries. Thus if the bottom-left is equal to the
				 * top-right, the rectangle is deemed to be empty. The
				 * target rectangle is "inclusive-inclusive"; that is, all
				 * boundaries are included in the rectangle.
				 * ...
				 * What does that meen?
				 * Image you want copy a rectangle of size (cx, cy) without
				 * resizing it. Let is left bottom point be at (0, 0). Then we
				 * have a rectangle similiar to the following:
				 *	                              cx - 1 
				 *     +------------------------------+ cy - 1
				 *     |                              |
				 *     |                              |
				 *     |                              |
				 *     |                              |
				 *     |                              |
				 *     |                              |
				 *     |                              |
				 *   0 +------------------------------+
				 *     0
				 * 
				 * TODO Erklrung fehlt noch!!
				 *
				\**************************************************************/

				memcpy( aptBmp, aptUpdate, 4 * sizeof(POINTL) );
				/* Query bitmap size and adjust rectangle where the image has
				 * to be drawn. */
				GpiQueryBitmapParameters( ptbmpCurrent->Image, &bmpInfo );

				if ( aptBmp[1].x > bmpInfo.cx ) {
					aptBmp[1].x = bmpInfo.cx - 1;
					aptBmp[3].x = bmpInfo.cx;
				}
				if ( aptBmp[1].y >= bmpInfo.cy ) {
					aptBmp[1].y = bmpInfo.cy - 1;
					aptBmp[3].y = bmpInfo.cy;
				}

#if 1
				GpiWCBitBlt( hpsBuffer, hbmpBkg, 4, aptUpdate, ROP_SRCCOPY, BBO_IGNORE );
				GpiWCBitBlt( hpsBuffer, ptbmpCurrent->Mask, 4, aptBmp, ROP_SRCAND, BBO_IGNORE );
				GpiWCBitBlt( hpsBuffer, ptbmpCurrent->Image, 4, aptBmp, ROP_SRCPAINT, BBO_IGNORE );
				/**************************************************************\
				 * Unlike GpiWCBitBlt the rectangles for GpiBitBlt are both
				 * noninclusive.
				 * (This funny inconsistency costed me some time !@%&*#?)
				\**************************************************************/
				aptUpdate[1].x++;
				aptUpdate[1].y++;
				GpiBitBlt( hps, hpsBuffer, 4, aptUpdate, ROP_SRCCOPY, BBO_IGNORE );
#else
				/* This methode is bad as it produces much flickering! */
				/* Draw background: */
				GpiWCBitBlt( hps, hbmpBkg, 4, aptUpdate, ROP_SRCCOPY, BBO_IGNORE );
				/* Mask out the sprite and Draw the sprite the image */
				GpiWCBitBlt( hpsBuffer, ptbmpCurrent->Mask, 4, aptBmp, ROP_SRCAND, BBO_IGNORE );
				GpiWCBitBlt( hpsBuffer, ptbmpCurrent->Image, 4, aptBmp, ROP_SRCPAINT, BBO_IGNORE );
#endif								
				WinEndPaint( hps );
        		return (MRESULT)0;
			} else {
				break;
			}
      } /* end case WM_PAINT */

      case WM_MOUSEMOVE: 
			if ( bInPullPhase ) {
				WinSetPointer( HWND_DESKTOP, hptrXhair );
				return (MRESULT)0;
			}
			break;

		case WM_BUTTON1DOWN: 
			if ( bInPullPhase ) {
				POINTL ptl = {	SHORT1FROMMP( mp1 ),	SHORT2FROMMP( mp1 ) };
				BITMAPINFOHEADER bmpInfo;

				GpiQueryBitmapParameters( ptbmpCurrent->Image, &bmpInfo );

				if ( ptl.x > 0 && ptl.y > 0 && ptl.x < bmpInfo.cx && ptl.y < bmpInfo.cy )
				{
					LONG lColor;

					GpiSetBitmap( hpsBuffer, ptbmpCurrent->Mask );
					lColor = GpiQueryPel( hpsBuffer, &ptl );
					GpiSetBitmap( hpsBuffer, hbmpBuffer );
					/* On different computers the background color of the two
					 * color bitmap  was not the same. (bug, feature or what???) */
					if ( (lColor != CLR_BLACK) && (lColor != CLR_BACKGROUND) ) {
						WinStopTimer( ProgInfo.hab, hwnd, tidPull );
						if ( ProgInfo.fPreload ) {
		   				ptbmpCurrent = &hbmpHit[0];
		   			} else {
		   				HPS hpsMem;
		   				BOOL bResult;
	
		   				bResult =
		   				GpiDeleteBitmap( ptbmpCurrent->Image );
		   				bResult&=
		   				GpiDeleteBitmap( ptbmpCurrent->Mask );
		   				if ( !bResult ) {
		   					WinSetCapture( HWND_DESKTOP, NULLHANDLE );
								PMFatal( ProgInfo.hab, "GpiDeleteBitmap" );
								return (MRESULT)0;
							}
	
		   				hpsMem = CreateMemPS( ProgInfo.hab );
		   				LoadHitImage( hpsMem, 0, ptbmpCurrent );
							DestroyMemPS( hpsMem );
						}
		   			iBmpIdxHit = 1;
			  			WinInvalidateRect( hwnd, NULL, FALSE );
						if ( ProgInfo.fSound ) {
							if ( sndBang.bInMem ) {
								AudioPlay( &sndBang, FALSE );
							} else {
								DosBeep( 100, 1 );
								DosBeep( 380, 1 );
								DosBeep( 110, 1 );
								DosBeep( 866, 1 );
							}
			     			WinStartTimer( ProgInfo.hab, hwnd, tidHit, 0 );
							if ( sndArrgh.bInMem ) {
								AudioPlayAsync( hwnd, &sndArrgh, TRUE );
							}
						} else {
			     		}
		     			bInPullPhase = FALSE;
		     		} /* end if lColor */
		     	} else {
					if ( ProgInfo.fSound ) {
						if ( sndBang.bInMem ) {
							AudioPlayAsync( hwnd, &sndBang, FALSE );
						} else {
							DosBeep( 100, 1 );
							DosBeep( 380, 1 );
							DosBeep( 110, 1 );
							DosBeep( 866, 1 );
						}
					}
	     		} /* end pt inside bitmap */
				return (MRESULT)0;
			}
			break;		
		/* end case WM_BUTTON1DOWN */

      case WM_TIMER: {
      	const ULONG tidTimer = (ULONG)mp1;

      	switch ( tidTimer ) { 
      		case tidSleep: {
      			const HWND hwndFrame = WinQueryWindow( hwnd, QW_PARENT );
      			const ULONG cxScreen =
						WinQuerySysValue( HWND_DESKTOP, SV_CXSCREEN );
      			const ULONG	cyScreen =
						WinQuerySysValue( HWND_DESKTOP, SV_CYSCREEN );
      			RECTL rect;

      			/* If the mouse is captured wait one period longer.
      			 * As a consequence only one guy can apear at the same time
      			 * but mouse capturing is a very critical call and I don't
      			 * want to disturb any serious application. */
      			if ( WinQueryCapture( HWND_DESKTOP ) ) {
      				return (MRESULT)0;
      			} 

		  			WinSetCapture( HWND_DESKTOP, hwnd );

      			/* Query window extents */
      			WinQueryWindowRect( hwndFrame, &rect );

      			/* Stop this timer */
     				WinStopTimer( ProgInfo.hab, hwnd, tidSleep );
      			/* Make the window visible and place it in front of all others */
      			WinSetWindowPos(
						hwndFrame, HWND_TOP,
						rand() % (cxScreen - rect.xRight),	/* Random position */
						rand() % (cyScreen - rect.yTop ),
						0, 0,											/* Don't change size */
						SWP_MOVE | SWP_SHOW | SWP_ZORDER );
					/* Save the background of the window */
      			hbmpBkg = GetWindowBitmap( hwnd, NULL );
      			if ( !hbmpBkg ) {
   					WinSetCapture( HWND_DESKTOP, NULLHANDLE );
						PMFatal( ProgInfo.hab, "GetWindowBitmap" );
						return( MRESULT)0;
      			}
	   			ptbmpCurrent = &hbmpPull[0];
					if ( !ProgInfo.fPreload ) {
	   				const HPS hpsMem = CreateMemPS( ProgInfo.hab );
	   				LoadPullImage( hpsMem, 0, ptbmpCurrent );
						DestroyMemPS( hpsMem );
					}
 	   			iBmpIdxPull = 1;
     				bInPullPhase = TRUE;
		  			WinInvalidateRect( hwnd, NULL, FALSE );
					WinUpdateWindow( hwnd );
					/* Make noise if wished */
					if ( ProgInfo.fSound && sndHaltStop.bInMem ) {
						/* Don't close the sound device after this sound. */
						AudioPlayAsync( hwnd, &sndHaltStop, FALSE );
					}
					/* Start the next timer. The first frame appears a little bit
					 * longer than the others. */
      			WinStartTimer(	ProgInfo.hab, hwnd,
						tidPull, 3 * ProgInfo.ulFrameRate / 2 );

      			
		     		return (MRESULT)0;
      		} /* end case tidSleep */

      		case tidMinutes:
      			/* This is a work around for the WinStartTimer bug */
      			if ( sSleep2Ct-- <= 0) {
      				WinStopTimer( ProgInfo.hab, hwnd, tidMinutes );
	     				WinStartTimer( ProgInfo.hab, hwnd, tidSleep,
   	  					(( ProgInfo.ulMinSleep 
     						+ 1000 * rand()
							% ( ProgInfo.ulMaxSleep - ProgInfo.ulMinSleep ) )
							& 0xFFFF ) );
     				}
					return (MRESULT)0;      				

     			case tidPull: {
     				/* During development I got some nasty errors because of the
     				 * following problem:
     				 * While processing this message the timer elapsed once again.
     				 * Because iBmpIdx already reached the highest possible value
     				 * the next tidPull message resulted in a value adressing
     				 * the hbmpBuffer variable. The program continued using this
     				 * bitmap handle and therefore produced errors when accessing
     				 * the GpiDeleteBitmap function as this it is protected due
     				 * to its permanent selection in hpsBuffer.
					 * The solution is rather trivial: Set iBmpIdx to zero
					 * if this timer should not be processed any more. */
     				if ( iBmpIdxPull == 0 ) {
     					return (MRESULT)0;
     				}
     				if ( ProgInfo.fPreload ) {
		   			ptbmpCurrent = &hbmpPull[iBmpIdxPull];
		   		} else {
	   				HPS hpsMem;
	   				BOOL bResult;

	   				bResult = 
	   				GpiDeleteBitmap( ptbmpCurrent->Image );
	   				bResult&= 
	   				GpiDeleteBitmap( ptbmpCurrent->Mask );
   					if ( !bResult ) {
	      				WinStopTimer( ProgInfo.hab, hwnd, tidPull );
   						WinSetCapture( HWND_DESKTOP, NULLHANDLE );
							PMFatal( ProgInfo.hab, "GpiDeleteBitmap" );
				     		return (MRESULT)0;
						}
	   				
	   				hpsMem = CreateMemPS( ProgInfo.hab );
	   				LoadPullImage( hpsMem, iBmpIdxPull, ptbmpCurrent );
						DestroyMemPS( hpsMem );
					}
		  			WinInvalidateRect( hwnd, NULL, FALSE );

	   			iBmpIdxPull++;
      			if ( iBmpIdxPull >= PULL_PICS  ) {
      				WinStopTimer( ProgInfo.hab, hwnd, tidPull );
      				iBmpIdxPull = 0;
      				bInPullPhase = FALSE;
	      			WinStartTimer( ProgInfo.hab, hwnd, tidFired, ProgInfo.ulFrameRate );
      			}
					return (MRESULT)0;
      		} /* end case tidPull */

      		case tidHit: {
      			/* Same problem as describe in case tidPull */
     				if ( iBmpIdxHit == 0 ) {
     					return (MRESULT)0;
     				}
     				if ( ProgInfo.fPreload ) {
		   			ptbmpCurrent = &hbmpHit[iBmpIdxHit];
		   		} else {
	   				HPS hpsMem;
	   				BOOL bResult;

	   				bResult =
	   				GpiDeleteBitmap( ptbmpCurrent->Image );
	   				bResult&=
	   				GpiDeleteBitmap( ptbmpCurrent->Mask );
   					if ( !bResult ) {
	      				WinStopTimer( ProgInfo.hab, hwnd, tidHit );
   						WinSetCapture( HWND_DESKTOP, NULLHANDLE );
							PMFatal( ProgInfo.hab, "GpiDeleteBitmap" );
				     		return (MRESULT)0;
						}

	   				hpsMem = CreateMemPS( ProgInfo.hab );
	   				LoadHitImage( hpsMem, iBmpIdxHit, ptbmpCurrent );
						DestroyMemPS( hpsMem );
					}
	   			iBmpIdxHit++;
		  			WinInvalidateRect( hwnd, NULL, FALSE );
      			if ( iBmpIdxHit >= HIT_PICS  ) {
      				iBmpIdxHit = 0;
      				WinStopTimer( ProgInfo.hab, hwnd, tidHit );
	      			WinStartTimer( ProgInfo.hab, hwnd, tidWait, 2 * ProgInfo.ulFrameRate );
						bRestart = (BOOL)ProgInfo.fRestart;
	      		}
		     		return (MRESULT)0;
      		} /* end case tidHit */

      		case tidFired:
      			/* Stop this timer */
     				WinStopTimer( ProgInfo.hab, hwnd, tidFired );
     				/* Make some noise */
					if ( ProgInfo.fSound ) {
						if ( sndBang.bInMem ) {
							AudioPlayAsync( hwnd, &sndBang, TRUE );
						} else {
							DosBeep( 100, 1 );
							DosBeep( 380, 1 );
							DosBeep( 110, 1 );
							DosBeep( 866, 1 );
						}
					}
      			/* Switch to last picture before shooting */
      			if ( ProgInfo.fPreload ) {
		   			ptbmpCurrent = &hbmpPull[PULL_PICS - 2];
		   		} else {
	   				HPS hpsMem;
	   				BOOL bResult;

	   				bResult =
	   				GpiDeleteBitmap( ptbmpCurrent->Image );
	   				bResult&=
	   				GpiDeleteBitmap( ptbmpCurrent->Mask );
   					if ( !bResult ) {
	   					WinSetCapture( HWND_DESKTOP, NULLHANDLE );
							PMFatal( ProgInfo.hab, "GpiDeleteBitmap" );
							return (MRESULT)0;
						}

	   				hpsMem = CreateMemPS( ProgInfo.hab );
	   				LoadPullImage( hpsMem, PULL_PICS - 2, ptbmpCurrent );
						DestroyMemPS( hpsMem );
					}
		  			WinInvalidateRect( hwnd, NULL, FALSE );
					bRestart = TRUE;
      			WinStartTimer( ProgInfo.hab, hwnd, tidWait, 4 * ProgInfo.ulFrameRate );
		     		return (MRESULT)0;
      		/* end case tidFired */
      		
      		case tidWait: {

     				WinStopTimer( ProgInfo.hab, hwnd, tidWait );
     				/* realease mouse capture */
		  			WinSetCapture( HWND_DESKTOP, NULLHANDLE );
      			WinShowWindow( WinQueryWindow( hwnd, QW_PARENT ), FALSE );
	   			if ( bRestart ) {
						BOOL bResult;

	   				bResult =
	   				GpiDeleteBitmap( hbmpBkg );
						if ( !ProgInfo.fPreload ) {
							bResult&=
		   				GpiDeleteBitmap( ptbmpCurrent->Image );
		   				bResult&=
		   				GpiDeleteBitmap( ptbmpCurrent->Mask );
		   			}
	   				if ( !bResult ) {
	   					PMFatal( ProgInfo.hab, "GpiDeleteBitmap" );
	   					return (MRESULT)0;
	   				}
		   			ptbmpCurrent = NULL;
		   			WinPostMsg( hwnd, MYM_STARTSLEEPTIMER, (MPARAM)0, (MPARAM)0 );
     				} else {
     					WinSendMsg( hwnd, WM_CLOSE, (MPARAM)0, (MPARAM)0 );
     				}
		     		return (MRESULT)0;
      		} /* end case tidWait */

      		default:
      			return WinDefWindowProc( hwnd, msg, mp1, mp2 );
     		} /* end switch tidTimer */
     	} /* end case WM_TIMER */
      		
		case WM_COMMAND: {
			/* No menus so far */
			break;
		} /* end case WM_COMMAND */
		
		case WM_CREATE: {
			const HAB hab = WinQueryAnchorBlock( hwnd );
			HPS hpsMem;
			HINI hini;
			SIZEL sizeWin = { 0, 0 };
			SIZEL sizeStart;
			LONG alCaps[2];
			ULONG ulHelp;
			BITMAPINFOHEADER bmpih;
			BITMAPINFOHEADER2 bmpih2;
			INT i;

			/* Load sounds if enabled */
			if ( ProgInfo.fSound ) {
				const ULONG ulSize =
					sizeof( SNDLOAD_THREADDATA ) + 3 * sizeof( PSOUNDINFO );
				PSNDLOAD_THREADDATA pSndLoadData;

				Xmalloc( pSndLoadData, ulSize, return (MRESULT)TRUE );
				pSndLoadData->ulStructSize = ulSize;
				pSndLoadData->hwndCaller = hwnd;
				pSndLoadData->apSndInfo[0] = &sndHaltStop;
				pSndLoadData->apSndInfo[1] = &sndGudenDag;
				pSndLoadData->apSndInfo[2] = &sndBang;
				pSndLoadData->apSndInfo[3] = &sndArrgh;
				if ( _beginthread(&LoadSounds, NULL, 0x8000, pSndLoadData) == -1 ) {
					ErrMsgRunLib( "_beginthread error ignored" );
					ProgInfo.fSound = 0;
					WinPostMsg( hwnd, MYM_STARTSLEEPTIMER, (MPARAM)0, (MPARAM)0 );
				}
			} else {
				/* To start the sleep timer we have to post this message to
				 * owerself */
				WinPostMsg( hwnd, MYM_STARTSLEEPTIMER, (MPARAM)0, (MPARAM)0 );
			}
					
			

			/* Load profile data */
			hini = PrfOpenProfile( hab, ProgInfo.pszIniName );			
			if ( hini == NULLHANDLE ) {
				/* PrfOpenProfile failed --> load all bitmaps to determine
				 * window size */
				ProgInfo.fPreload = 1;
			} else {
				ULONG ulNumRead = sizeof( SIZEL );
				PrfQueryProfileData( hini,
					ProgInfo.pszAppTitle, PrfKeyNames.pszWindowSize,
					&sizeWin, &ulNumRead );
				PrfCloseProfile( hini );
				if ( ulNumRead != sizeof( SIZEL ) || !( sizeWin.cx && sizeWin.cy) ) {
					ProgInfo.fPreload = 1;
				}
			} /* end else !hini */
			/* We will rewrite window size profile data if the window size
			 * changes during installation. */
			sizeStart = sizeWin;
			
			/* Load resource data */
			hptrXhair = WinLoadPointer( HWND_DESKTOP, NULLHANDLE, IDPTR );
			if ( !hptrXhair ) {
					ErrMsgPM( hab, "WinLoadPointer" );
					return (MRESULT)TRUE;
			}

			hpsMem = CreateMemPS( hab );
			if ( hpsMem == NULLHANDLE ) {
				ErrMsgPM( hab, "CreateMemPS" );
				return (MRESULT)TRUE;
			}

			if ( ProgInfo.fPreload ) {
				/* Load bitmap files */
				for ( i = 0; i < PULL_BMPS; i++ ) {
					if ( !LoadPullImage( hpsMem, i, &hbmpPull[i] ) ) {
						ErrMsgPM( hab, "LoadTransparentBitmap" );
						DestroyMemPS( hpsMem );
						return (MRESULT)TRUE;
					}
					/* Note: GpiQueryBitmapDimesion doesn't work as it
					 * returns values defined earlier by GpiSetBitmapDimesion only. */
					GpiQueryBitmapParameters( hbmpPull[i].Image, &bmpih );
					if ( sizeWin.cx < bmpih.cx ) sizeWin.cx = bmpih.cx;
					if ( sizeWin.cy < bmpih.cy ) sizeWin.cy = bmpih.cy;
				}
	
				for ( i = 0; i < HIT_BMPS; i++ ) {
					if ( !LoadHitImage( hpsMem, i, &hbmpHit[i] ) ) {
						ErrMsgPM( hab, "LoadTransparentBitmap" );
						DestroyMemPS( hpsMem );
						return (MRESULT)TRUE;
					}
					/* Be careful: GpiQueryBitmapDimesion doesn't work as it
					 * returns values defined earlier by GpiSetBitmapDimesion only. */
					GpiQueryBitmapParameters( hbmpHit[i].Image, &bmpih );
					if ( sizeWin.cx < bmpih.cx ) sizeWin.cx = bmpih.cx;
					if ( sizeWin.cy < bmpih.cy ) sizeWin.cy = bmpih.cy;
				}
			} else {
#if 0
				/* This part has been wiped out because it was only used
				 * to determine the window size. Now I use profile functions
				 * which results in a better startup performance. */
				HBITMAP hbmpDummy;
				BOOL bResult = TRUE;
				
				for ( i = 0; bResult && (i < PULL_BMPS); i++ ) {
					
#if LOAD_BMPS_FROM_FILE
					hbmpDummy = LoadBitmapFile( hpsMem, apszPullFiles[i] );
#else
					hbmpDummy = GpiLoadBitmap( hpsMem, NULLHANDLE, alPullIds[i], 0, 0 );
#endif
					bResult = 1 && hbmpDummy;
					bResult&=
					GpiQueryBitmapParameters( hbmpDummy, &bmpih );
					if ( sizeWin.cx < bmpih.cx ) sizeWin.cx = bmpih.cx;
					if ( sizeWin.cy < bmpih.cy ) sizeWin.cy = bmpih.cy;
					bResult&=
					GpiDeleteBitmap( hbmpDummy );
				}
				
				for ( i = 0; bResult && (i < HIT_BMPS); i++ ) {
#if LOAD_BMPS_FROM_FILE
					hbmpDummy = LoadBitmapFile( hpsMem, apszHitFiles[i] );
#else
					hbmpDummy = GpiLoadBitmap( hpsMem, NULLHANDLE, alHitIds[i], 0, 0 );
#endif					
					bResult = 1 && hbmpDummy;
					bResult&=
					GpiQueryBitmapParameters( hbmpDummy, &bmpih );
					if ( sizeWin.cx < bmpih.cx ) sizeWin.cx = bmpih.cx;
					if ( sizeWin.cy < bmpih.cy ) sizeWin.cy = bmpih.cy;
					bResult&=
					GpiDeleteBitmap( hbmpDummy );
				}
				if ( !bResult ) {
					DestroyMemPS( hpsMem );
					return (MRESULT)TRUE;
				}
#endif
				memset( hbmpPull, 0, PULL_BMPS * sizeof( TRANSPARENT_BITMAP ) );
				memset( hbmpHit, 0, HIT_BMPS * sizeof( TRANSPARENT_BITMAP ) );
			} /* end else ProgInfo.fPreload */
			if ( (sizeWin.cx != sizeStart.cx) || (sizeWin.cy != sizeStart.cy) ) {
				hini = PrfOpenProfile( hab, ProgInfo.pszIniName );			
				if ( hini != NULLHANDLE ) {
					PrfWriteProfileData( hini,
						ProgInfo.pszAppTitle, PrfKeyNames.pszWindowSize,
						&sizeWin, sizeof( SIZEL ) );
					PrfCloseProfile( hini );
					}
				} /* end if !hini */

			/* Adjust the size of the frame window so that every bitmap fits
			 * inside. */
			WinSetWindowPos( WinQueryWindow( hwnd, QW_PARENT ), NULLHANDLE,
				0, 0,	sizeWin.cx, sizeWin.cy, SWP_SIZE );
			/* We don't destroy hpsMem but use it as a buffer when redrawing
			 * bitmaps */
			DevQueryCaps( GpiQueryDevice( hpsMem ), CAPS_COLOR_PLANES, 2, alCaps );
			bmpih2.cbFix = 16;
		 	bmpih2.cx = sizeWin.cx;
			bmpih2.cy = sizeWin.cy;
			bmpih2.cPlanes = alCaps[0];
			bmpih2.cBitCount = alCaps[1];
			hbmpBuffer = GpiCreateBitmap( hpsMem, &bmpih2, 0, NULL, NULL );
			if ( !hbmpBuffer ) {
				ErrMsgPM( hab, "GpiCreateBitmap" );
				DestroyMemPS( hpsMem );
				return (MRESULT)TRUE;
			}
			GpiSetBitmap( hpsMem, hbmpBuffer );

			srand( (unsigned int) time( &ulHelp ) );
#if 0
			/* The timer isn't started here anymore because I want to load all
			 * sounds first. Therefore the timer is started when we recieve the
			 * MYM_ENDSNDLOAD message */
			WinPostMsg( hwnd, MYM_STARTSLEEPTIMER, (MPARAM)0, (MPARAM)0 );
#endif			

			GpiSetColor( hpsMem, CLR_WHITE );
			GpiSetBackColor( hpsMem, CLR_BLACK );
			hpsBuffer = hpsMem;
			break;
		} /* end case WM_CREATE */

		case WM_DESTROY: {
			/* Bitmaps are destroyed automaticly when the program terminates but
			 * I have had some very funny effects if I killed a program that
			 * still had a presentation space in use! */
			if ( hpsBuffer ) {
				GpiSetBitmap( hpsBuffer, NULLHANDLE );
				DestroyMemPS( hpsBuffer );
			}
			break;
		} /* end case WM_DESTROY */

		case MYM_ENDSNDLOAD: {
			if ( mp1 ) {			
				free( (PVOID)mp1 );
				if ( sndGudenDag.bInMem ) {
					AudioPlay( &sndGudenDag, TRUE );
					AudioFree( &sndGudenDag );
				}
			}
			/* Posting this message is a bit redundant as we could simply
			 * remove it and the following return statement. However I prefer
			 * this way. */
			WinPostMsg( hwnd, MYM_STARTSLEEPTIMER, (MPARAM)0, (MPARAM)0 );
			return (MRESULT)0;
		} /* end case MYM_ENDSNDLOAD */

		case MYM_STARTSLEEPTIMER: {
			ULONG ulHelp;
			ULONG ulTimeOut;

			/* Start the `minutes' timer */
			ulHelp = ProgInfo.ulMinSleep
				+ 1000 * rand() % ( ProgInfo.ulMaxSleep - ProgInfo.ulMinSleep );
			ulTimeOut = ulHelp >> 16;
			sSleep2Ct = ulHelp && 0xFFFF;
			if ( !WinStartTimer( ProgInfo.hab, hwnd, tidMinutes, ulTimeOut ) ) {
				ErrMsgPM( ProgInfo.hab, "WinStartTimer" );
			}
			return (MRESULT)0;
		} /* end case MYM_STARTSLEEPTIMER */
	} /* end switch msg */
	if ( !AudioWndProc( hwnd, msg, mp1, mp2 ) ) {
		return WinDefWindowProc( hwnd, msg, mp1, mp2 );
	}
	return (MRESULT)0;
}

/*************************************************************************\
 * FUNCTION main
 * Do some initialisation and loop through messages.
\*************************************************************************/
INT main( INT argc, CHAR *argv[] )
{
	HMQ hmq = NULLHANDLE;
	QMSG qmsg;							/* message-queue of our program			*/
	ULONG fWndCtrlData;				/* window style for the created win		*/
	BOOL fRegistered; 				/* return value by various functions	*/
	HWND hwndClient;
	CHAR pszClientClass[] = "shootem.child";

	if ( (ProgInfo.hab = WinInitialize(0)) == NULLHANDLE ) {
		fputs( "WinInitialize failed\n ", stderr );
		DosBeep( 440, 100 );
		exit( 1 );
	}

	if ( !(hmq = WinCreateMsgQueue( ProgInfo.hab, 0 )) ) {
		fputs( "WinCreateMsgQueue failed\n ", stderr );
		DosBeep( 440, 100 );
		WinTerminate( ProgInfo.hab );
		exit( 1 );
	}

	if ( !InitProgInfo( argc, argv, "shootem" ) ) {
		WinDestroyMsgQueue( hmq );
		WinTerminate( ProgInfo.hab );
		return 1;
	}

	fRegistered = 
	WinRegisterClass(	ProgInfo.hab, pszClientClass, MainClientProc, 0, 0 );
	if ( !fRegistered ) {
		fputs( "WinRegisterClass failed\n ", stderr );
		DosBeep( 440, 100 );
		WinDestroyMsgQueue( hmq );
		WinTerminate( ProgInfo.hab );
		return 1;
	}

	fWndCtrlData = FCF_ICON				| FCF_AUTOICON		|
						FCF_TASKLIST;

	ProgInfo.hwndMain =
	WinCreateStdWindow(
		HWND_DESKTOP, 	 				/* Parent-window */
		WS_SAVEBITS,					/* don't make window visible yet	*/
		&fWndCtrlData,					/* window parameters, defined above */
		pszClientClass,				/* window class of client window */
		ProgInfo.pszAppTitle,		/* Title Bar text */
		0,									/* client-window style */
		0,				 					/* resources are bound within .exe */
		IDR_MAIN,						/* Frame-window identifier */
		&hwndClient );					/* output client-window handle */
	if ( !ProgInfo.hwndMain ) {
		PMFatal( ProgInfo.hab, "WinCreateStdWindow" );
		WinDestroyMsgQueue( hmq );
		WinTerminate( ProgInfo.hab );
		return 1;
	} /* endif */

	while( WinGetMsg( ProgInfo.hab, (PQMSG)&qmsg, NULLHANDLE, 0L, 0L ))
		 WinDispatchMsg( ProgInfo.hab, (PQMSG)&qmsg );

	if ( ProgInfo.fSaveSettings ) StoreSettings();

	/* clean up */
	if ( ProgInfo.hwndMain ) WinDestroyWindow( ProgInfo.hwndMain );
	if ( hmq ) WinDestroyMsgQueue( hmq );
	if ( ProgInfo.hab ) WinTerminate( ProgInfo.hab );
	return 0;
} /* end of main	 */

